#pragma once

/* Genode includes */
#include <util/misc_math.h>
#include <util/string.h>
#include <base/lock.h>
#include <base/rpc_client.h>
#include <base/attached_dataspace.h>
#include <tcb_info_session/tcb_info_session.h>

namespace Tcb_info
{
	using namespace Genode;

	using Genode::memcpy;

	class Session_client;
}


class Tcb_info::Session_client : public Rpc_client<Session>
{
	private:

		Lock _lock { };

		/**
		 * Shared-memory buffer used for carrying the payload
		 * of read/write operations
		 */
		Attached_dataspace _io_buffer;

	public:

		Session_client(Region_map &local_rm, Capability<Session> cap) :
			Rpc_client<Session> { cap },
			_io_buffer          { local_rm, call<Rpc_dataspace>() }
		{ }


		Transaction_handle begin_transaction(Tcb_type type,
		                                     const Security_domain& security_domain) override
		{
			if (security_domain.length() > _io_buffer.size())
				throw Tcb_info::Buffer_too_small { };

			Lock::Guard guard(_lock);

			memcpy(_io_buffer.local_addr<char>(), security_domain.string(), security_domain.length());

			return call<Rpc_begin_transaction>(type, security_domain.length());
		}


		size_t get_tcb_info_hash(Transaction_handle transaction_handle,
		                         uint8_t* hash_buffer, size_t hash_buffer_size) override
		{
			if ((hash_buffer == nullptr) || (hash_buffer_size == 0))
				throw Tcb_info::Buffer_too_small { };

			Lock::Guard _guard(_lock);

			size_t reply_bytes = call<Rpc_get_tcb_info_hash>(transaction_handle);

			if (reply_bytes > _io_buffer.size()) throw Tcb_info::Buffer_too_small { };
			if (reply_bytes > hash_buffer_size) throw Tcb_info::Buffer_too_small { };

			memcpy(hash_buffer, _io_buffer.local_addr<char>(), reply_bytes);

			return reply_bytes;
		}


		size_t get_tcb_info_tree(Transaction_handle transaction_handle,
		                         uint8_t* tree_buffer, size_t tree_buffer_size) override
		{
			if ((tree_buffer == nullptr) || (tree_buffer_size == 0))
				throw Tcb_info::Buffer_too_small { };

			Lock::Guard _guard(_lock);

			size_t reply_bytes = call<Rpc_get_tcb_info_tree>(transaction_handle);

			if (reply_bytes > _io_buffer.size()) throw Tcb_info::Buffer_too_small { };
			if (reply_bytes > tree_buffer_size) throw Tcb_info::Buffer_too_small { };

			memcpy(tree_buffer, _io_buffer.local_addr<char>(), reply_bytes);

			return reply_bytes;
		}

		void provide_additional_hash(const char* hash_buffer, size_t hash_size,
	                                 const char* name_buffer, size_t name_size) override
		{
			if ((hash_buffer == nullptr) && (hash_size != 0))
				throw Tcb_info::Buffer_too_small { };

			if ((name_buffer == nullptr) && (name_size != 0))
				throw Tcb_info::Buffer_too_small { };

			if ((hash_size+name_size) > _io_buffer.size())
				throw Tcb_info::Buffer_too_small { };

			Lock::Guard _guard(_lock);

			memcpy(_io_buffer.local_addr<char>(),           hash_buffer, hash_size);
			memcpy(_io_buffer.local_addr<char>()+hash_size, name_buffer, name_size);

			call<Rpc_provide_additional_hash>(hash_size, name_size);
		}
};
